/********************************************************************\

  Name:         frontend.c
  Created by:   Stefan Ritt

  Contents:     Experiment specific readout code (user part) of
                Midas frontend. This example simulates a "trigger
                event" and a "scaler event" which are filled with
                CAMAC or random data. The trigger event is filled 
                with two banks (ADC0 and TDC0), the scaler event 
                with one bank (SCLR).

  $Log: frontend.cc,v $
  Revision 1.1.1.1  2002/08/13 12:48:12  kensh


  Start version.

  Revision 1.13  2000/09/28 13:28:24  midas
  Added manually triggered events

  Revision 1.11  2000/08/21 10:32:51  midas
  Added max_event_size, set event_buffer_size = 10*max_event_size

  Revision 1.10  2000/03/13 18:53:29  pierre
  - Added 2nd arg in readout functions (offset for Super event)

  Revision 1.9  2000/03/02 22:00:00  midas
  Added number of subevents as zero

  Revision 1.8  1999/02/24 16:27:01  midas
  Added some "real" readout code

  Revision 1.7  1999/01/20 09:03:38  midas
  Added LAM_SOURCE_CRATE and LAM_SOURCE_STATION macros

  Revision 1.6  1999/01/19 10:27:30  midas
  Use new LAM_SOURCE and LAM_STATION macros

  Revision 1.5  1998/11/09 09:14:41  midas
  Added code to simulate random data

  Revision 1.4  1998/10/29 14:27:46  midas
  Added note about FE_ERR_HW in frontend_init()

  Revision 1.3  1998/10/28 15:50:58  midas
  Changed lam to DWORD

  Revision 1.2  1998/10/12 12:18:58  midas
  Added Log tag in header


\********************************************************************/

#include <iostream>
#include <iomanip>
#include <stdio.h>
#include <stdlib.h>
#include "midas.h"
#include "mcstd.h"

#include <unistd.h>
#include <time.h>
#include "ngf_1885n.h"

static int    debugLevel = 0;
extern bool   PollTrigger;

//////

/* make frontend functions callable from the C framework */
#ifdef __cplusplus
extern "C" {
#endif

/*-- Globals -------------------------------------------------------*/

/* The frontend name (client name) as seen by other MIDAS clients   */
char *frontend_name = "Frontend";
/* The frontend file name, don't change it */
char *frontend_file_name = __FILE__;

/* frontend_loop is called periodically if this variable is TRUE    */
BOOL frontend_call_loop = FALSE;

/* a frontend status page is displayed with this frequency in ms */
INT display_period = 4000;

/* maximum event size produced by this frontend */
INT max_event_size = 10000000; // 10MB

/* maximum event size for fragmented events (EQ_FRAGMENTED) */
INT max_event_size_frag = 10000000;

/* buffer size to hold events */
INT event_buffer_size = 2 * 10000000; // 20MB

/*-- Function declarations -----------------------------------------*/

INT frontend_init();
INT frontend_exit();
INT begin_of_run(INT run_number, char *error);
INT end_of_run(INT run_number, char *error);
INT pause_run(INT run_number, char *error);
INT resume_run(INT run_number, char *error);
INT frontend_loop();

INT poll_event(INT source, INT count, BOOL test);
INT interrupt_configure(INT cmd, INT source, PTYPE adr);

INT read_trigger_event(char *pevent, INT off);
INT read_scaler_event(char *pevent, INT off);

/*-- Equipment list ------------------------------------------------*/

#undef USE_INT

EQUIPMENT equipment[] = {

  { "NGF-ADC",            /* equipment name */
    1, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
#ifdef USE_INT
    EQ_INTERRUPT,         /* equipment type */
#else
    EQ_POLLED,            /* equipment type */
#endif
    LAM_SOURCE(0,0xFFFFFF),/* event source crate 0, all stations */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    //RO_RUNNING |          /* read only when running */
    //RO_ODB,               /* and update ODB */ 
    RO_RUNNING,           /* read only when running */
    500,                  /* poll for 500ms */
    0,                    /* stop run after this event limit */
    0,                    /* number of sub events */
    0,                    /* don't log history */
    "", "", "",
    read_trigger_event,   /* readout routine */
  },

  { "Scaler",             /* equipment name */
    2, 0,                 /* event ID, trigger mask */
    "SYSTEM",             /* event buffer */
    EQ_PERIODIC | 
    EQ_MANUAL_TRIG,       /* equipment type */
    0,                    /* event source */
    "MIDAS",              /* format */
    TRUE,                 /* enabled */
    //RO_RUNNING |
    //RO_TRANSITIONS |      /* read when running and on transitions */
    //RO_ODB,               /* and update ODB */ 
    RO_RUNNING |
    RO_TRANSITIONS,       /* read when running and on transitions */
    10000,                /* read every 10 sec */
    0,                    /* stop run after this event limit */
    0,                    /* number of sub events */
    0,                    /* log history */
    "", "", "",
    read_scaler_event,    /* readout routine */
  },

  { "" }
};

#ifdef __cplusplus
}
#endif

/********************************************************************\
              Callback routines for system transitions

  These routines are called whenever a system transition like start/
  stop of a run occurs. The routines are called on the following
  occations:

  frontend_init:  When the frontend program is started. This routine
                  should initialize the hardware.
  
  frontend_exit:  When the frontend program is shut down. Can be used
                  to releas any locked resources like memory, commu-
                  nications ports etc.

  begin_of_run:   When a new run is started. Clear scalers, open
                  rungates, etc.

  end_of_run:     Called on a request to stop a run. Can send 
                  end-of-run event and close run gates.

  pause_run:      When a run is paused. Should disable trigger events.

  resume_run:     When a run is resumed. Should enable trigger events.

\********************************************************************/

/*-- Frontend Init -------------------------------------------------*/

INT frontend_init()
{
  if ( debugLevel )
    std::cerr << "frontend_init" << std::endl;

  /* hardware initialization */
  std::cout << std::endl;

  // check host
  char  host[256];
  gethostname( host, 256 );
  if ( strcmp( host,"kldaq2" ) != 0 ){
    std::cerr << std::endl << std::endl << std::endl 
              << "This program is for kldaq2 : this host=" << host
              << std::endl << std::endl << std::endl; 
    abort();
  }

  // Init NGF and Fastbus
  initNgf();
  
  // Command List
  initNgfCommandList();

  sleep(3);
  
  /* print message and return FE_ERR_HW if frontend should not be started */

  return SUCCESS;
}

/*-- Frontend Exit -------------------------------------------------*/

INT frontend_exit()
{
  if ( debugLevel )
    std::cerr << "frontend_exit" << std::endl;

  // terminate NGF
  termNgf();
  
  return SUCCESS;
}

/*-- Begin of Run --------------------------------------------------*/

INT begin_of_run(INT run_number, char *error)
{
  if ( debugLevel )
    std::cerr << "begin_of_run" << std::endl;

  /* put here clear scalers etc. */

  // Start NGF sequence
  beginRunNgf();
  
  return SUCCESS;
}

/*-- End of Run ----------------------------------------------------*/

INT end_of_run(INT run_number, char *error)
{
  if ( debugLevel )
    std::cerr << "end_of_run" << std::endl;
 
  // Stop NGF Sequence
  endRunNgf();
  
  return SUCCESS;
}

/*-- Pause Run -----------------------------------------------------*/

INT pause_run(INT run_number, char *error)
{
  if ( debugLevel )
    std::cerr << "pause_run" << std::endl;

  // Stop NGF Sequence
  pauseRunNgf();

  return SUCCESS;
}

/*-- Resuem Run ----------------------------------------------------*/

INT resume_run(INT run_number, char *error)
{
  if ( debugLevel )
    std::cerr << "resume_run" << std::endl;

  // Start NGF sequence
  resumeRunNgf();
  
  return SUCCESS;
}

/*-- Frontend Loop -------------------------------------------------*/

INT frontend_loop()
{
  if ( debugLevel )
    std::cerr << "frontend_loop" << std::endl;

  /* if frontend_call_loop is true, this routine gets called when
     the frontend is idle or once between every event */
  return SUCCESS;
}

/*------------------------------------------------------------------*/

/********************************************************************\
  
  Readout routines for different events

\********************************************************************/

/*-- Trigger event routines ----------------------------------------*/

INT poll_event(INT source, INT count, BOOL test)
/* Polling routine for events. Returns TRUE if event
   is available. If test equals TRUE, don't return. The test
   flag is used to time the polling */
{
  if ( debugLevel )
    std::cerr << "poll_event : count=" << count << " test=" << test
              <<  std::endl;
  
  // Wait Trigger Interrupt
  for( int i=0; i<=count; i++ ){
    if ( interruptSource(1) && !test ){
      return( true );
    }
    else if ( interruptSource(2) && !test ){
      bufferEventNgf();
      return( 0 );
    }
    else{
      static const struct timespec ts = { 0, 10000000 }; // 10 msec
      if ( !PollTrigger ){
        nanosleep( &ts,NULL );
      }
    }
  }
  
  if ( listFailedNgf() ){
    if ( !test ){
      extern Ngf  ngf;
      std::cerr << "poll_event : NGF list mode failed and restarted. Seq=0x" 
                << std::setw(8) << std::setfill('0') << std::hex
                << ngf.SeqStatusReg()
                << " FB=0x" 
                << ngf.FbStatusReg1()
                << ",0x" 
                << ngf.FbStatusReg2()
                << std::dec << std::endl;
      resetListNgf();
    }
  }
  
  if ( debugLevel ){
    extern Ngf  ngf;
    std::cerr << "poll_event : SeqStat=" 
              << std::setw(8) << std::setfill('0') << std::hex
              << ngf.SeqStatusReg()
              << " Fifo=" 
              << ngf.SeqFifoAndEclNimInputReg()
              << std::dec << std::endl;
  }
  
  return 0;
}

/*-- Interrupt configuration ---------------------------------------*/

INT interrupt_configure(INT cmd, INT source, PTYPE adr)
{
  if ( debugLevel )
    std::cerr << "interrupt_configure" << std::endl;

  switch(cmd)
    {
    case CMD_INTERRUPT_ENABLE:
      if ( debugLevel )
        std::cerr << "interrupt_configure:CMD_INTERRUPT_ENABLE" << std::endl;
      break;

    case CMD_INTERRUPT_DISABLE:
      if ( debugLevel )
        std::cerr << "interrupt_configure:CMD_INTERRUPT_DISABLE" << std::endl;
      break;

    case CMD_INTERRUPT_ATTACH:
      if ( debugLevel )
        std::cerr << "interrupt_configure:CMD_INTERRUPT_ATTACH" << std::endl;
      // Enable Interrupt for SPILL signal
      //sigset( SIGINT,(void(*)(int))adr );
      break;

    case CMD_INTERRUPT_DETACH:
      if ( debugLevel )
        std::cerr << "interrupt_configure:CMD_INTERRUPT_DETACH" << std::endl;
      break;
    }
  return SUCCESS;
}

/*-- Event readout -------------------------------------------------*/

INT read_trigger_event(char *pevent, INT off)
{
  if ( debugLevel )
    std::cerr << "read_trigger_event" << std::endl;
  
  /* init bank structure */
  bk_init32(pevent);
  
  // fill bank
  fillBankNgf( pevent );

  return bk_size(pevent);
}

/*-- Scaler event --------------------------------------------------*/

INT read_scaler_event(char *pevent, INT off)
{
  if ( debugLevel )
    std::cerr << "read_scaler_event" << std::endl;

  //
  DWORD *pdata;
  
  /* init bank structure */
  bk_init32(pevent);
  
  /* create SCLR bank */
  bk_create(pevent, "SCLR", TID_DWORD, &pdata);
  
  /* read scaler bank */
  
  /* close SCLR bank */
  bk_close(pevent, pdata);
  
  return bk_size(pevent);
}


